from pydantic import BaseModel, validator, Extra
import typing
from pydantic.error_wrappers import ValidationError

from pydantic.errors import ArbitraryTypeError


try:
    from typing import Literal
except ImportError:
    # backport
    from typing_extensions import Literal


class CommandTypes:
    SUB_COMMAND = 1
    SUB_COMMAND_GROUP = 2
    STRING = 3
    INTEGER = 4
    BOOLEAN = 5
    USER = 6
    CHANNEL = 7
    ROLE = 8


class CommandChoice(BaseModel):
    name: str
    value: str


class CommandOption(BaseModel):
    class Config:
        arbitrary_types_allowed = True
        extra = Extra.allow

    name: str
    description: str
    type: int
    required: bool = False
    choices: typing.Optional[
        typing.Union[typing.List[dict], typing.List[CommandChoice]]
    ] = []

    # options: typing.Optional[typing.Union[typing.List[CommandChoice]]]

    # @validator("options")
    # def options_allow_only_if_subcommand(cls, v):
    #    if cls.type != 1:
    #        raise ValidationError("Type must be 1 in order to have options.")
    #    return v


class SubcommandOption(BaseModel):
    class Config:
        arbitrary_types_allowed = True

    name: str
    description: str
    type: Literal[2] = 2
    options: typing.List[CommandOption]

    @validator("options")
    def options_must_contain_type_1(cls, v):  # pylint: disable=no-self-argument
        item: CommandOption
        for item_location, item in enumerate(v):
            if item.type != 1:
                raise ValidationError(
                    f"CommandOptions located in <{item_location}> must be have type of 1 due to parent being a subcommand."
                )
        return v


class DiscordCommand(BaseModel):
    name: str
    description: str
    options: typing.List[typing.Union[SubcommandOption, CommandOption]]



from fastapi import APIRouter, Request, Response
from loguru import logger
from .middlewares.verification import DiscordVerificationMiddleware
from .models.incoming import IncomingDiscordInteraction
from .eventer import EventHandler
from .response import DiscordResponse, NotReadyResponse
import json
import typing

router = APIRouter()
interaction = EventHandler()


_RAISE_FOR_TESTING = False


@router.post("/interactions")
async def handle_interactions(request: Request) -> Response:
    logger.info("interaction recieved.")

    _get_request_body = json.loads(request.state._cached_body.decode())
    logger.info(_get_request_body)
    if _get_request_body["type"] == 1:
        logger.info("handling ACK Ping.")
        return {"type": 1}

    _parse_to_object = IncomingDiscordInteraction(**_get_request_body)

    if interaction.check_event_exists(_parse_to_object.data.name) == False:
        logger.debug("discarding event not existing.")
        return {}

    if len(_parse_to_object.data.options) > 0:
        arguments = {x.name: x.value for x in _parse_to_object.data.options}
    else:
        arguments = {}

    arguments["ctx"] = _parse_to_object

    # Check the type hint for the return type, fallback for checking the type if no hints are provided
    try:
        _type_hinted_request = interaction.view_event_function_return_type(
            _parse_to_object.data.name
        )
        _type_hinted_returned_value = _type_hinted_request["return"]
        if _type_hinted_returned_value == DiscordResponse:
            _get_res = await interaction.emit(_parse_to_object.data.name, **arguments)
            return _get_res.response
        elif _type_hinted_returned_value == NotReadyResponse:
            return None
        elif _type_hinted_returned_value == dict:
            return await interaction.emit(_parse_to_object.data.name, **arguments)
    except KeyError:
        logger.error(
            "unable to find return value for type hint.. resorting to guessing.."
        )
        if _RAISE_FOR_TESTING == True:
            raise AssertionError("No hinting!")
    except Exception:
        logger.exception("unhandled exception for returning hinted value")
        raise

    interaction_data = await interaction.emit(_parse_to_object.data.name, **arguments)
    if isinstance(interaction_data, DiscordResponse):
        interaction_data: DiscordResponse
        return interaction_data.response

    if isinstance(interaction_data, dict):
        return interaction_data

    if isinstance(interaction_data, NotReadyResponse):
        return None


from pydantic import BaseModel
from .helper.embed import Embed
import typing
from .errors.network import DiscordAPIError
import httpx

try:
    from typing import Literal
except ImportError:
    # backport
    from typing_extensions import Literal

if typing.TYPE_CHECKING:
    from .main import Dispike
    from .models import IncomingDiscordInteraction


class DiscordResponse(object):
    def __init__(
        self,
        content: str = None,
        tts: bool = False,
        embeds: typing.List[Embed] = [],
        show_user_input: bool = False,
    ):

        if content != None:
            if isinstance(content, str) == False:
                raise TypeError(f"Content must be a string")
            elif content == "":
                content = None

        # if isinstance(content, str) == False or content == "" or content != None:
        #    raise TypeError(f"content must be a string. recieved: {content}")

        if isinstance(tts, bool) == False:
            raise TypeError("tts must be a bool")

        self._content = content
        self._tts = tts
        self._embeds = [x.to_dict() for x in embeds]
        if show_user_input == False:
            self._type_response = 3
        else:
            self._type_response = 4

    @property
    def embeds(self):
        return self._embeds

    def add_new_embed(self, embed_to_add: Embed):
        if isinstance(embed_to_add, Embed):
            self._embeds.append(embed_to_add.to_dict())
        else:
            raise TypeError("embed must be a Embed object.")

    @property
    def content(self):
        return self._content

    @content.setter
    def content(self, new_content_string: str):
        self._content = new_content_string

    @property
    def tts(self):
        return self._tts

    @tts.setter
    def tts(self, new_tts: bool):
        self._tts = new_tts

    @property
    def response(self):
        if self.content == "":
            self.content = None

        return {
            "type": self._type_response,
            "data": {"tts": self.tts, "content": self.content, "embeds": self.embeds},
        }

    def __call__(self):
        return self.response


class NotReadyResponse(object):
    """This is just a representation for a response that is not ready.
    You probably want to use this if you are not ready to immediately return
    a response to an incoming interaction

    You have access to the original arugmnents that arrived with the interaction by accessing
    the attribute ``.args``, this returns a dict of kwargs.

    Lower-level access is available by accessing ``._interaction_context``.

    Attributes:
        - sync_send_callback
        - async_send_callback
        - args


    """

    def __init__(
        self, bot: "Dispike", interaction_context: "IncomingDiscordInteraction"
    ):
        """Incoming context to have a response later.

        Args:
            bot (Dispike): Incoming bot object
            interaction_context (IncomingDiscordInteraction): incoming discord interaction
        """
        self._application_id = bot._application_id
        self._interaction_id = interaction_context.id
        self._interaction_token = interaction_context.token

        if len(interaction_context.data.options) > 0:
            self.args = {x.name: x.value for x in interaction_context.data.options}
        else:
            self.args = {}

        self._interaction_context = interaction_context

    def sync_send_callback(self, ready_response: DiscordResponse):
        """Send a callback synchronously

        Args:
            ready_response (DiscordResponse): The DiscordResponse that has been created.

        Raises:
            DiscordAPIError: Any issues that may arrive

        Returns:
            True: if everything works out
        """
        try:
            send_callback = httpx.post(
                f"https://discord.com/api/v8/interactions/{self._interaction_id}/{self._interaction_token}/callback",
                json=ready_response.response,
            )
            if send_callback.status_code not in [200, 201]:
                raise DiscordAPIError(send_callback.status_code, send_callback.text)
            return True
        except Exception:
            raise

    async def async_send_callback(self, ready_response: DiscordResponse):
        """Send a callback asynchronously

        Args:
            ready_response (DiscordResponse): The DiscordResponse that has been created.

        Raises:
            DiscordAPIError: Any issues that may arrive

        Returns:
            True: if everything works out
        """
        try:
            send_callback = await httpx.AsyncClient().post(
                f"https://discord.com/api/v8/interactions/{self._interaction_id}/{self._interaction_token}/callback",
                json=ready_response.response,
            )
            if send_callback.status_code not in [200, 201]:
                raise DiscordAPIError(send_callback.status_code, send_callback.text)
            return True
        except Exception:
            raise


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dispike',
 'dispike.errors',
 'dispike.helper',
 'dispike.middlewares',
 'dispike.models',
 'dispike.models.discord_types',
 'dispike.register',
 'dispike.register.models']

package_data = \
{'': ['*']}

install_requires = \
['PyNaCl>=1.4.0,<2.0.0',
 'fastapi>=0.63.0,<0.64.0',
 'httpx>=0.16.1,<0.17.0',
 'loguru>=0.5.3,<0.6.0',
 'pydantic>=1.7.3,<2.0.0',
 'typing-extensions>=3.7.4,<4.0.0',
 'uvicorn>=0.13.2,<0.14.0']

setup_kwargs = {
    'name': 'dispike',
    'version': '0.5.5a0',
    'description': 'library for interacting with discord slash commands via an independently hosted server. Powered by FastAPI',
    'long_description': '# dispike\n\n***\n[![codecov](https://codecov.io/gh/ms7m/dispike/branch/master/graph/badge.svg?token=E5AXLZDP9O)](https://codecov.io/gh/ms7m/dispike) ![Test Dispike](https://github.com/ms7m/dispike/workflows/Test%20Dispike/badge.svg?branch=master) ![PyPi Link](https://img.shields.io/badge/Available%20on%20PyPi-Dispike-blue?logo=pypi&link=%22https://pypi.org/project/dispike%22) ![PyPiVersion](https://img.shields.io/badge/dynamic/json?color=blue&label=PyPi%20Version&query=%24.info.version&url=https%3A%2F%2Fpypi.org%2Fpypi%2Fdispike%2Fjson)\n\n***\n\n\n\nan *extremely-extremely* early WIP library for easily creating REST-based webhook bots for discord using the new Slash Commands feature. \n\nPowered by FastAPI.\n\n\n\n## Example (a simple bot)\n\n```python\n\n# Incoming models use this.\nfrom dispike.models.incoming import IncomingDiscordInteraction # For Type Hinting\n\n# Main Dispike Object\nfrom dispike import Dispike\n\n# For creating commands.\nfrom dispike.register.models import DiscordCommand, CommandOption, CommandChoice, CommandTypes\n\n# For responding to commands.\nfrom dispike.response import DiscordResponse\n\n\nbot = Dispike(\n  client_public_key="< PublicKey >"\n  bot_token="< Bot Token >",\n  application_id="< AppID >"\n)\n\n\n\n# Let\'s register a command.\n# /wave <person>\n# Learn more about registering commands in the Discord Docs.\n\ncommand_to_be_created = DiscordCommand(\n    name="wave" # this is the main command name.,\n    description="Send a wave to a nice person! 👋 ",\n    options=[\n        CommandOption(\n            name="person" # this is the attribute assigned to the value passed.,\n            description="person to target" # this describes the value to pass,\n          \trequired=True,\n            type=CommandTypes.USER\n        )\n    ]\n)\n\n# Nice, we created a command. /wave <person to ping>. Let\'s register it.\n\nbot.register(\n\tcommand_to_be_created\n)\n\n\n# We\'ve registered, now it\'s time to build the bot! \n# Arguments that you pass to your function are \n# the same arguments you defined/registered with discord..\n# Don\'t forget to have an argument for a ctx (context). \n\n@bot.interaction.on("wave")\nasync def handle_send_wave(person: int, ctx: IncomingDiscordInteraction):\n  logger.info("Recieved a wave!")\n  \n\n  # this is what we will be returning. Let\'s edit it.\n  # Discord will pass the user id..\n  _response = DiscordResponse()\n  _response.content f"👋 Hi @<{person}>."\n  \n  return _response.response\n  \n# Run your bot.\n# Powered by uvicorn.\n\nbot.run(port=5000)\n    \n```\nEmbeds are also available (Copied from Discord.py)\n\n```python\nfrom dispike.helper import Embed\n\n_response = DiscordResponse()\n_response.embed = Embed() # your embed created.\n```\n\n## Getting, Deleting and Editing Commands\n\n\n\n```python\nfrom dispike import Dispike\n\n# For creating commands.\nfrom dispike.register.models import DiscordCommand, CommandOption, CommandChoice, CommandTypes\n\n\nbot = Dispike(\n\tclient_public_key="< PublicKey >"\n  bot_token="< Bot Token >",\n  application_id="< AppID >"\n)\n\nall_commands = bot.get_commands()\n\nfor command in all_commands:\n  print(f"{command.name} -> {command.id}")\n \ntarget_command = all_commands[0]\n\ncommand_to_be_created = DiscordCommand(\n    name="salute" # this is the main command name.,\n    description="Send a salute to a nice person! ",\n    options=[\n        CommandOption(\n            name="person" # this is the attribute assigned to the value passed.,\n            description="person to target" # this describes the value to pass,\n          \trequired=True,\n            type=CommandTypes.USER\n        )\n    ]\n)\n\nbot.edit_command(target_command.id, command_to_be_created)\nbot.delete_command(all_commands[1].id)\n```\n\n\n\nWhen configuring your endpoint on discord settings, be sure to append ``/interactions`` to your domain.\n\n<p >\n    <img\n      alt="Website"\n      src="./docs/images/domain.png"\n    />\n</p>\n\n## Caveats\n\n- ~~Does not handle registring new commands.~~\n- ~~Does not handle anything other then string responses. (However you are free to return any valid dict in your handler.)~~\n- ~~Not on PyPi~~\n- Does not speak over the discord gateway. You\'ll need a server to handle requests and responses.\n- Python 3.6+\n- Does not support the following endpoints\n  - [Create Followup Message](https://discord.com/developers/docs/interactions/slash-commands#create-followup-message)\n  - [Edit Followup Message](https://discord.com/developers/docs/interactions/slash-commands#edit-followup-message)\n  - [Delete Followup Message](https://discord.com/developers/docs/interactions/slash-commands#delete-followup-message)\n- Handling followup messages.\n\n\n\n# Development\n\nHelp is wanted in mantaining this library. Please try to direct PRs to the ``dev`` branch, and use black formatting (if possible).\n\n![Test Dispike](https://github.com/ms7m/dispike/workflows/Test%20Dispike/badge.svg?branch=dev)\n',
    'author': 'Mustafa Mohamed',
    'author_email': 'ms7mohamed@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ms7m/dispike',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
